////////////////////////////////////////////////////////////////
// Flash: Intel G18 x16||x16 version
//
//  Flash write- and erase-routines optimized specifically for Intel G18
//  StrataFlash.   And implemented and tested for only one member of
//  the family, at that.
//
//  If you want to know all about this API, see the large, complete
//  comment atop the accomanying header-file.
//
////////////////////////////////////////////////////////////////

#include <io.h>
#include <stdio.h>
#include "system.h"

//EPCQ Serial Flash Controller Register Offset
#define RD_STATUS_OFFSET 0x0 // R, Status Register, contains the block protection bits that represent the memory sector to be protected from write or erase
#define RD_SID_OFFSET 0x4 // R, invalid here as JTAG Avalon MM interface will run into crash once it try to access this register
#define RD_RDID_OFFSET 0x8 // R
#define MEM_OP_OFFSET 0xC // Write Only
#define ISR_OFFSET 0x10 // WR
#define IMR_OFFSET 0x14 // WR
#define CHIP_SELECT_OFFSET 0x18 // Write Only

#define FLASH_BLOCK_SIZE 0x10000  //  64KB
#define BYTES_PER_WORD 4



int flash_qspi_write(alt_u32 base, alt_u32 offset, alt_u32* buffer, int word_count)
{
    int sectorIndex = (int) ((offset & 0xFFFF0000) >> 16);
    int sectorOffset = (int) offset & 0xFFFF;
    int count = 0;
	int status;
	int cnt = 10;

    int error_code = 0;


    while (word_count > 0) {
        // Figure-out if we're going to run past the next erase-block-boundary
		int chunk_size = FLASH_BLOCK_SIZE / BYTES_PER_WORD; // number of times we will write to flash (= 160 byte_count/4 = 40)
		//printf("ODD OFFSET CASE: offset= 0x%x, word_count= 0x%x\n",offset, word_count);


		if (offset % FLASH_BLOCK_SIZE != 0) { // offset is in bytes
			// Make this chunk only big-enough to get us to the next multiple of FLASH1616_MAX_BYTE_SIZE.
			int next_even_offset = ((offset / FLASH_BLOCK_SIZE) + 1 ) * FLASH_BLOCK_SIZE;
			int bytes_to_write	 = next_even_offset - offset; // offset if a byte-y thing
#ifdef MAN_DEBUG
			printf("ODD OFFSET CASE: offset= 0x%x, word_count= 0x%x, next_even_offset= 0x%x bytes_to_write= 0x%x\n",
				  offset, word_count, next_even_offset, bytes_to_write);
#endif

			// We write 32-bit ganged 16 bit half-words to the flash...not bytes.
			chunk_size = bytes_to_write / BYTES_PER_WORD;
		}

		// Deal with the "we're at the end" case.
		if (chunk_size > word_count)  // if maximum is greater than what's left
		{
			chunk_size = word_count; // just write however many are needed for what's left
#ifdef MAN_DEBUG
			printf("END_CASE: offset= 0x%x, word_count= 0x%x, 0x%x\n", offset, word_count);
#endif
		}

        //qspiMemWriteWords((int) offset, buffer, chunk_size);
		int i = 0;
        for (i = 0; i < chunk_size; i++) {
        	IOWR_32DIRECT(EXT_FLASH_AVL_MEM_BASE, offset + i * 4, *(buffer + i));
        }

		do {
			OSTimeDlyHMSM(0, 0, 0, 200);
			status = IORD_32DIRECT(EXT_FLASH_AVL_CSR_BASE, RD_STATUS_OFFSET);
		} while ((1 == (status & 0x1)) && (0 < (cnt--)));

		if (1 == (status & 0x1)) {
			printf("Write timeout, Status is 0x%08x", status);
			return status;
		}

        buffer += chunk_size; // Buffer contains 32-bit WORDS
        offset += chunk_size * BYTES_PER_WORD; // Offset given in 8-bit BYTES
        word_count -= chunk_size; // 0x20a0000 --> 0x20a00a0 (correct)
    }
    return 0;
}


int flash_qspi_erase(alt_u32 base, alt_u32 offset, alt_u32 byte_count)
{

    int sectorIndex = (int) ((offset & 0xFFFF0000) >> 16);

    int numOfSector = (byte_count + 0xFFFF) / 0x10000;

    IOWR_32DIRECT(EXT_FLASH_AVL_CSR_BASE, ISR_OFFSET, 0x3);

    //Set all area as Unprptected
    IOWR_32DIRECT(EXT_FLASH_AVL_CSR_BASE, MEM_OP_OFFSET, 0x00001003);

    int i = 0;
    for (i = 0; i < numOfSector; i++) {
    	IOWR_32DIRECT(EXT_FLASH_AVL_CSR_BASE, MEM_OP_OFFSET, ((sectorIndex + i) & 0xFFFF) << 8 | 0x2);
    }

    int status;
    int cnt = 10;
    do {
    	OSTimeDlyHMSM(0, 0, 0, 200);
        status = IORD_32DIRECT(EXT_FLASH_AVL_CSR_BASE, RD_STATUS_OFFSET);
    } while ((1 == (status & 0x1)) && (0 < (cnt--)));

    if (1 == (status & 0x1)) {
        printf("Erase timeout, Status is 0x%08x", status);
        return status;
    }

	return 0;
}

